from dataclasses import dataclass

from torch import Tensor


@dataclass
class AudioData:
    """
    Output audio data for :class:`~.audio_reader.AudioReader`.

    Attributes:
        audio (:class:`~torch.Tensor`): audio float tensor of (C, T)
        sample_rate (``int``): sample rate of the audio

    """
    audio: Tensor
    sample_rate: int


import os
from typing import Callable, List, Dict, Iterable, Sequence, Union, Any, Tuple

import numpy
import numpy as np
import torch
from fn import F, _, Stream
from fn.op import identity
from fn.uniform import reduce
from torch import Tensor
from torch.nn import ModuleList, Module

from .type import T, A


def reduce_dict_by(key: str, op: Callable[[T, T], T]) -> Callable[[List[Dict[str, T]]], T]:
    """
    Apply reducer function to a list of dictionaries.

    Args:
        key (``str``): The key of the value in dicts
        op (``(T, T) -> T``): A reduce operation function used to reduce the values.

    Returns:
        ``List[Dict[str, T]] -> T``: The reducer aggregation function for the list of dictionaries.

    Examples::

        >>> x = [
        >>>     {"a": 1, "b": 2, "c": torch.Tensor([3])},
        >>>     {"a": 3, "b": 4, "c": torch.Tensor([5])},
        >>>     {"a": 2.3, "b": -1, "c": torch.Tensor([0])}
        >>> ]
        >>> reduce_dict_by("a", lambda x1, x2: x1 + x2)(x)
        6.3
        >>> reduce_dict_by("b", lambda x1, x2: x1 - x2)(x)
        -1
        >>> reduce_dict_by("c", lambda x1, x2: 10 * x1 + x2)(x)
        tensor([350.])

    """
    return F() >> (map, _[key]) >> list >> F(reduce, op)


def summarize_dict_by(key: str, op: Callable[[Union[Sequence[T], T]], T]
) -> Callable[[List[Dict[str, T]]], T]:
    """
    Apply summarizer function to a list of dictionaries.

    Args:
        key (``str``): The key of the value in dicts
        op (``(Sequence[T] | T) -> T``):
            A summarizer operation function used to summarize value from a sequence.

    Returns:
        ``List[Dict[str, T]] -> T``:
            The summarizer aggregation function for the list of dictionaries.

    Examples::

        >>> x = [
        >>>    {"a": 1, "b": torch.Tensor([2]), "c": 3, "d": np.array([1.5])},
        >>>    {"a": 3, "b": torch.Tensor([4]), "c": 5, "d": np.array([2.5])},
        >>>    {"a": 2.3, "b": torch.Tensor([-1]), "c": 0, "d": np.array([-1.0])}
        >>> ]
        >>> summarize_dict_by("a", sum)(x)
        6.3
        >>> summarize_dict_by("b", torch.mean)(x)
        tensor(1.6667)
        >>> def some_func(x):
        >>>     x = list(map(str, x))
        >>>     x = "".join(x)
        >>>     x = int(x)
        >>>     return x
        >>> summarize_dict_by("c", some_func)(x)
        350
        >>> summarize_dict_by("d", lambda x: np.sum(x, axis=0))(x)
        array([3.])

    """
    return F() >> (map, _[key]) >> list >> ifelse(
        lambda xs: type(xs[0]) is Tensor,
        func_true=torch.vstack,
        func_false=ifelse(
            lambda xs: type(xs[0]) is np.ndarray,
            func_true=numpy.vstack,
            func_false=_
        )
    ) >> op


def generate_inf_seq(items: Iterable[Any]) -> Stream:
    """
    Generate an infinity late-evaluate sequence.

    Args:
        items [``Iterable[Any]``]: Original items.

    Returns:
        :class:`~fn.stream.Stream`: The infinity sequence.

    Examples::

        >>> seq = generate_inf_seq(["a", 1, ["e0", "e1"]])
        >>> list(seq[:10])
        ['a', 1, ['e0', 'e1'], 'a', 1, ['e0', 'e1'], 'a', 1, ['e0', 'e1'], 'a']

    Note:
        The repeated reference type objects are the same object.

        >>> a_list = ["a", "b"]
        >>> seq = list(generate_inf_seq([a_list])[:3])
        >>> seq
        [['a', 'b'], ['a', 'b'], ['a', 'b']]
        >>> a_list.append("c")
        >>> seq
        [['a', 'b', 'c'], ['a', 'b', 'c'], ['a', 'b', 'c']]

    """
    s = Stream()
    return s << items << map(_, s)


def compose(fs: Union[ModuleList, Sequence[Callable]]) -> F:
    """
    Compose functions as a pipeline function.

    Args:
        fs (``Sequence[Callable]`` | :class:`~torch.nn.ModuleList`): The functions input for composition.

    Returns:
        :class:`~fn.func.F`: The composed output function.

    Examples::

        >>> f = lambda x: x + 1
        >>> g = lambda x: x * 2
        >>> h = lambda x: x ** 2
        >>> x = 1
        >>> h(g(f(x))) == compose([f, g, h])(x)
        True

    """
    return reduce(_ >> _, fs, F())


def listdir(path: str) -> List[str]:
    """
    The full path version of :func:`os.listdir`.

    Get the all listdir result with input path.

    Args:
        path (``str``): The path for listdir.

    Returns:
        ``List[str]``: listdir result with input path.

    Examples::

        >>> os.listdir("tensorneko/util")[:3]  # python os library
        ['configuration.py', 'func.py', 'string_getter.py']
        >>> listdir("tensorneko/util")[:3]  # tensorneko
        ['tensorneko/util/configuration.py',
         'tensorneko/util/func.py',
         'tensorneko/util/string_getter.py']

    """
    files = filter(_ != ".DS_Store", os.listdir(path))
    return list(map(F(os.path.join, path), files))


def with_printed(x: Any, func: Callable = identity) -> Any:
    """
    An identity function but with printed to console with some transform.

    Args:
        x (``Any``): Input.
        func (``Callable``, optional): A function used to apply the input for printing.

    Returns:
        ``Any``: Identity output.

    Examples::

        >>> x = torch.tensor([1.5, 2.5, 3.5])
        >>> y = with_printed(x, lambda tensor: tensor.dtype)
        torch.float32
        >>> x == y
        tensor([True, True, True])

    """
    print(func(x))
    return x


def with_printed_shape(x: A) -> A:
    """
    An identity function but with shape printed .

    Args:
        x (:class:`~torch.Tensor` | :class:`numpy.ndarray`): Input.

    Returns:
        :class:`~torch.Tensor` | :class:`numpy.ndarray`: Identity output.

    Examples::

        >>> x = torch.tensor([1.5, 2.5, 3.5])
        >>> y = with_printed_shape(x)
        torch.Size([3])
        >>> x == y
        tensor([True, True, True])
    """
    return F(with_printed, func=lambda tensor: tensor.shape)(x)


def ifelse(predicate: Callable[[Any], bool], func_true: Callable, func_false: Callable) -> Callable:
    """
    A function composition util for if-else control flow.

    Args:
        predicate (``(...) -> bool``): A predicate produce a bool.
        func_true (``(...) -> Any``): If the bool from predicate is True, return this function.
        func_false (``(...) -> Any``): If the bool from predicate is False, return this function.

    Returns:
        ``(...) -> Any``: Composed function with if-else control flow.
    """

    def wrapper(*args, **kwargs):
        if predicate(*args, **kwargs):
            return func_true(*args, **kwargs)
        else:
            return func_false(*args, **kwargs)

    return wrapper


def is_bad_num(x: Tensor) -> Tensor:
    """
    Checking the element in input tensor is bad number (NaN or Inf).

    Args:
        x (:class:`~torch.Tensor`): Input tensor.

    Returns:
        :class:`~torch.Tensor`: Checking result for bad numbers.
    """
    return torch.logical_or(torch.isnan(x), torch.isinf(x))


def count_parameters(module: Module) -> int:
    """
    Counts the number of parameters of a :class:`~torch.nn.Module`.

    Args:
        module (:class:`~torch.nn.Module`): The input module for counting parameters.

    Returns:
        ``int``: Number of parameters.
    """
    return sum(p.numel() for p in module.parameters())


def dict_add(*dicts: dict):
    """
    Merge multiple dictionaries.

    Args:
        *dicts:

    Returns:

    """
    new_dict = {}
    for each in dicts:
        new_dict.update(each)
    return new_dict


def as_list(*args, **kwargs) -> list:
    """
    Returns:
        ``list``: The list of args.

    Examples::

        >>> as_list(1, 2, 3, 4, key=5)
        [1, 2, 3, 4, 5]
    """
    return list(args) + list(kwargs.values())


from typing import Optional, Union, Iterable, Sequence, Callable

from torch import Tensor
from torch.nn import ModuleList

from ..neko_module import NekoModule
from ..layer import Linear
from ..util import generate_inf_seq, ModuleFactory, compose


class MLP(NekoModule):
    """
    The MLP module is a sequential module with multiple linear layers.

    The activation, normalization and dropout are only applied to hidden layers.

    Args:
        neurons (``Sequence[int]``): The list of neurons for the MLP.

        bias (``bool | Iterable[bool]``, optional): The bias option of Linear layers. Default ``True``.

        build_activation (``() -> Module | Iterable[() -> Module]``, optional): The activation builder for linear
            layers. Specify single module builder will apply to all linear layers except the final one.
            Default ``None``.

        build_normalization (``() -> Module | Iterable[() -> Module]``, optional): The normalization builder for linear
            layers. Specify single module builder will apply to all linear layers except the final one.
            Default ``None``.

        normalization_after_activation (``bool``, optional): Set True then applying normalization after activation.
            Default ``False``.

        dropout_rate (``float``, optional): The dropout rate for this linear layer. 0 means no dropout applied.
            Default ``0``.

    Attributes:
        layers (:class:`~torch.nn.ModuleList`): The module list of :class:`~tensorneko.layer.Linear` layers.

    Examples:

        Create a MLP module with 4 layers, with neurons 768 -> 1024 -> 512 -> 10.
        Contains ReLU, BatchNorm and 0.5 Dropout.

        .. code-block:: python

            mlp = tensorneko.module.MLP(
                neurons=[784, 1024, 512, 10],
                build_activation=torch.nn.ReLU,
                build_normalization=[
                    lambda: torch.nn.BatchNorm1d(1024),
                    lambda: torch.nn.BatchNorm1d(512)
                ],
                dropout_rate=0.5
            )

    """

    def __init__(self, neurons: Sequence[int], bias: Union[bool, Iterable[bool]] = True,
        build_activation: Optional[Union[ModuleFactory, Iterable[ModuleFactory]]] = None,
        build_normalization: Optional[Union[ModuleFactory, Iterable[ModuleFactory]]] = None,
        normalization_after_activation: bool = False, dropout_rate: float = 0.
    ):
        super().__init__()
        n_features = neurons[1:]

        if type(bias) is bool:
            bias = generate_inf_seq([bias])

        if isinstance(build_activation, Callable) or build_activation is None:
            build_activation = generate_inf_seq([build_activation])
        if isinstance(build_normalization, Callable) or build_normalization is None:
            build_normalization = generate_inf_seq([build_normalization])

        self.layers: ModuleList[Linear] = ModuleList(
            [Linear(neurons[i], neurons[i + 1], bias[i], build_activation[i], build_normalization[i],
                normalization_after_activation, dropout_rate
            ) for i in range(len(n_features) - 1)
            ] + [
                Linear(neurons[-2], neurons[-1], bias[len(neurons) - 2])
            ]
        )

    def forward(self, x: Tensor) -> Tensor:
        f = compose(self.layers)
        return f(x)



from pytorch_lightning import Callback, Trainer, LightningModule


class LrLogger(Callback):
    """Log learning rate in each epoch start."""

    def on_train_epoch_start(self, trainer: Trainer, pl_module: LightningModule) -> None:
        for i, optimizer in enumerate(trainer.optimizers):
            for j, params in enumerate(optimizer.param_groups):
                pl_module.log(f"opt{i}_lr{j}", params["lr"])



import torch
from torch import Tensor, reshape

from ..neko_module import NekoModule
from ..util import Shape


class Reshape(NekoModule):
    """
    The module version of :func:`torch.reshape` operation.

    Args:
        shape: (:class:`tensorneko.util.type.Shape`): The target shape of input tensor.

    Examples::

        >>> reshape_a = Reshape((2, 2))
        >>> a = torch.arange(4.)
        >>> reshape_a(a)
        tensor([[ 0.,  1.],
                [ 2.,  3.]])

        >>> reshape_b = Reshape((-1,))
        >>> b = torch.tensor([[0, 1], [2, 3]])
        >>> reshape_b(b)
        tensor([ 0,  1,  2,  3])

    """

    def __init__(self, shape: Shape):
        super().__init__()
        self.shape = shape

    def forward(self, x: Tensor) -> Tensor:
        return reshape(x, shape=self.shape)



import re
import unittest


class TestVersion(unittest.TestCase):

    def test_package_version(self):
        VERSION_PATTERN = r"""
            v?
            (?:
                (?:(?P<epoch>[0-9]+)!)?                           # epoch
                (?P<release>[0-9]+(?:\.[0-9]+)*)                  # release segment
                (?P<pre>                                          # pre-release
                    [-_\.]?
                    (?P<pre_l>(a|b|c|rc|alpha|beta|pre|preview))
                    [-_\.]?
                    (?P<pre_n>[0-9]+)?
                )?
                (?P<post>                                         # post release
                    (?:-(?P<post_n1>[0-9]+))
                    |
                    (?:
                        [-_\.]?
                        (?P<post_l>post|rev|r)
                        [-_\.]?
                        (?P<post_n2>[0-9]+)?
                    )
                )?
                (?P<dev>                                          # dev release
                    [-_\.]?
                    (?P<dev_l>dev)
                    [-_\.]?
                    (?P<dev_n>[0-9]+)?
                )?
            )
            (?:\+(?P<local>[a-z0-9]+(?:[-_\.][a-z0-9]+)*))?       # local version
        """

        regex = re.compile(
            r"^\s*" + VERSION_PATTERN + r"\s*$",
            re.VERBOSE | re.IGNORECASE,
        )

        # read setup.py file
        def read(file):
            with open(file, "r", encoding="UTF-8") as input_file:
                text = input_file.read()
            return text

        try:
            version = read("version.txt")
        except FileNotFoundError:
            version = read("../version.txt")

        self.assertTrue(regex.match(version) is not None)



import torch
from torch import Tensor, log

from ..neko_module import NekoModule


class Log(NekoModule):
    """
    The module version of :func:`torch.log` operation.

    Args:
        eps (``float``, optional): A bias applied to the input to avoid ``-inf``. Default ``0``.

    Examples::

        >>> log = Log()
        >>> a = torch.randn(5)
        >>> a
        tensor([ 2.3020, -0.8679, -0.2174,  2.4228, -1.2341])
        >>> log(a)
        tensor([0.8338,    nan,    nan, 0.8849,    nan])

    """

    def __init__(self, eps: float = 0.):
        super().__init__()
        self.eps: float = eps

    def forward(self, x: Tensor) -> Tensor:
        return log(x) if self.eps == 0 else log(x + self.eps)



from typing import Union

import torchaudio.backend.sox_io_backend
from torchaudio.backend.common import AudioMetaData

from .audio_data import AudioData


class AudioReader:
    """AudioReader for reading audio file"""

    @staticmethod
    def of(path: str, return_info: bool = False) -> Union[AudioData, AudioMetaData]:
        """
        Read audio tensor from given file.

        Args:
            path (``str``): Path to the audio file.
            return_info (``bool``, optional): True will return info rather than audio itself. Default ``False``.

        Returns:
            :class:`~.audio_data.AudioData` | :class:`~torchaudio.backend.common.AudioMetaData`:
                The Audio data in the file.
        """
        if return_info:
            return torchaudio.info(path)
        else:
            return AudioData(*torchaudio.load(path))



# TODO: need refactor

import torch
from torch import nn
import torch.nn.functional as F
from torch.nn.parameter import Parameter
import numpy as np
from torch.nn import init


def upscale_2d(x):
    return nn.functional.interpolate(x, scale_factor=2, mode='bilinear', align_corners=False)


def downscale_2d(x):
    # since the factor is same for height and width the bilinear downsampling is just like an avg pool
    return F.avg_pool2d(x, 2, 2)  # downscales in factor 2


def compute_r1_gradient_penalty(d_result_real, real_images):
    real_grads = torch.autograd.grad(d_result_real.sum(), real_images, create_graph=True, retain_graph=True)[0]
    # real_grads = torch.autograd.grad(d_result_real, real_images,
    #                                  grad_outputs=torch.ones_like(d_result_real),
    #                                  create_graph=True, retain_graph=True)[0]
    r1_penalty = 0.5 * torch.sum(real_grads.pow(2.0), dim=[1, 2, 3]) # Norm on all dims but batch

    return r1_penalty

class StyleInstanceNorm2d(nn.Module):
    def __init__(self, channels):
        super(StyleInstanceNorm2d, self).__init__()
        self.instance_norm = nn.InstanceNorm2d(channels, affine=False)

    def forward(self, x):
        # TODO: implement  nn.InstanceNorm2d here to save redundant mean/var computation
        assert(len(x.shape) == 4)
        m = torch.mean(x, dim=[2, 3], keepdim=True)
        std = torch.sqrt(torch.mean((x - m) ** 2, dim=[2, 3], keepdim=True))
        style = torch.cat((m, std), dim=1)

        x = self.instance_norm(x)

        return x, style


class LearnablePreScaleBlur(nn.Module):
    """
    From StyleFan paper:
    "We replace the nearest-neighbor up/downsampling in both net-works with bilinear sampling, which we implement by
    low-pass filtering the activations with a separable 2nd order bi-nomial filter after each upsampling layer and
    before each downsampling layer [Making Convolutional Networks Shift-Invariant Again]."
    """
    def __init__(self, channels):
        super(LearnablePreScaleBlur, self).__init__()
        f = np.array([[1, 2, 1]])
        f = np.matmul(f.transpose(), f)
        f = f / np.sum(f)
        kernel = torch.Tensor(f).view(1, 1, 3, 3).repeat(channels, 1, 1, 1)
        self.register_buffer('weight', kernel)
        self.groups = channels

    def forward(self, x):
        x = F.conv2d(x, weight=self.weight, groups=self.groups, padding=1)
        return x


def pixel_norm(x, epsilon=1e-8):
    return x / torch.rsqrt(torch.mean(x.pow(2.0), dim=1, keepdim=True) + epsilon)
    # return x * torch.rsqrt(torch.mean(x.pow(2.0), dim=1, keepdim=True) + epsilon)


class StyleAffineTransform(nn.Module):
    '''
    The 'A' unit in StyleGan:
    Learned affine transform A, this module is used to transform the midiate vector w into a style vector.
    it outputs a mean and std scalar for each channel for a given number of channels.
    this mean and std are used in AdaIn as facror and bias to shift the norm of the input channels
    '''
    def __init__(self, dim_latent, n_channel):
        super().__init__()
        self.transform = LREQ_FC_Layer(dim_latent, n_channel * 2)
        # "the biases associated with mean that we initialize to one"
        self.transform.bias.data[:n_channel] = 1
        self.transform.bias.data[n_channel:] = 0

    def forward(self, w):
        style = self.transform(w).unsqueeze(2).unsqueeze(3)
        return style


class NoiseScaler(nn.Module):
    '''
    Learned per-channel scale factor, used to scale the noise
    'B' unit in SttleGan
    '''

    def __init__(self, n_channel):
        super().__init__()
        # per channel wiehgt
        self.weight = nn.Parameter(torch.zeros((1, n_channel, 1, 1)))

    def forward(self, noise):
        result = noise * self.weight
        return result


class AdaIn(nn.Module):
    '''
    adaptive instance normalization
    Shift the mean and variance of each chnnel in the input image by a given factor and bias
    '''

    def __init__(self, n_channel):
        super().__init__()
        self.norm = nn.InstanceNorm2d(n_channel)

    def forward(self, image, style):
        factor, bias = style.chunk(2, 1) # split number of chenels to two
        normed_input = self.norm(image)
        rescaled_image = normed_input * factor + bias
        return rescaled_image


class LREQ_FC_Layer(nn.Linear):
    def __init__(self, in_features, out_features, bias=True):
        super(LREQ_FC_Layer, self).__init__(in_features, out_features, bias)
        init.normal_(self.weight)
        if bias:
            init.zeros_(self.bias)
        self.c = np.sqrt(2.0) / np.sqrt(in_features)

    def forward(self, input):
        return F.linear(input,
                        self.weight * self.c,
                        self.bias)



class Lreq_Conv2d(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size, padding=1, stride=1, output_padding=0, dilation=1, bias=True):
        super(Lreq_Conv2d, self).__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = (kernel_size, kernel_size)
        self.stride = (stride, stride)
        self.padding = (padding, padding)
        self.output_padding = (output_padding, output_padding)
        self.dilation = (dilation, dilation)

        self.weight = Parameter(torch.Tensor(out_channels, in_channels, *self.kernel_size))
        init.normal_(self.weight)
        if bias:
            self.bias = Parameter(torch.Tensor(out_channels))
            init.zeros_(self.bias)
        else:
            self.register_parameter('bias', None)

        self.c = np.sqrt(2.0) / np.sqrt(np.prod(self.kernel_size) * in_channels)

    def __str__(self):
        return f"LreqConv2d({self.in_channels}, {self.out_channels}, k={self.kernel_size[0]}, p={self.padding[0]})"


    def forward(self, x):
        return F.conv2d(x,
                        self.weight * self.c,
                        self.bias,
                        stride=self.stride,
                        padding=self.padding,
                        dilation=self.dilation
                        )


from torchvision.io import read_video

from .video_data import VideoData


class VideoReader:
    """VideoReader for reading video file"""

    @staticmethod
    def of(path: str) -> VideoData:
        """
        Read video tensor from given file.

        Args:
            path: Path to the video file.

        Returns:
            :class:`~torch.Tensor`: A float tensor of video (T, C, H, W), with value range of 0. to 1.
        """
        video, audio, info = read_video(path)
        video = video.permute(0, 3, 1, 2) / 255
        return VideoData(video, audio, info)



from einops.layers.torch import Rearrange as _Rearrange

from .func import reduce_dict_by, summarize_dict_by, generate_inf_seq, compose, listdir, with_printed, \
    with_printed_shape, is_bad_num, ifelse, dict_add, count_parameters, as_list
from .type import ModuleFactory, Shape, Device
from .configuration import Configuration
from .string_getter import get_activation, get_loss
from .pipe import Args as __

from fn.underscore import shortcut
from fn import func

Rearrange = _Rearrange
_ = shortcut
F = func.F


__all__ = [
    "reduce_dict_by",
    "summarize_dict_by",
    "generate_inf_seq",
    "compose",
    "listdir",
    "with_printed",
    "with_printed_shape",
    "is_bad_num",
    "ifelse",
    "dict_add",
    "count_parameters",
    "as_list",
    "ModuleFactory",
    "Shape",
    "Device",
    "Configuration",
    "get_activation",
    "get_loss",
    "__",
    "Rearrange",
    "_",
    "F"
]


from typing import Union, Optional

from einops.layers.torch import Rearrange
from fn import F
from torch import Tensor
from torch.nn import Conv2d

from ..neko_module import NekoModule
from ..util import ModuleFactory, Shape


class Patching(NekoModule):

    def __init__(self):
        super().__init__()
        # TODO
        raise NotImplementedError

    def forward(self, x: Tensor) -> Tensor:
        return x


class PatchEmbedding2d(NekoModule):
    """
    The patch embedding layer from Vision Transformer (ViT) (Dosovitskiy, et al., 2020). This layer will take a image
    as input, and divide to patches, and project to a provided number of channels.

    Here it is implemented by a equivalent convolution layer with same stride length and kernel size.

    Args:
        image_size (:class:`~tensorneko.util.type.Shape`): The size of input shape. (C, H, W).

        patch_size (``int`` | :class:`~tensorneko.util.type.Shape`): The side length of patch or a patch Shape of
            (H, W).

        embedding (``int``): Output embedding dimension.

        strides (``int`` | :class:`~tensorneko.util.type.Shape`, optional): The strides of the patch window.
            Default is the size of patch.

        build_normalization (``() -> torch.nn.Module``, optional): An normalization module builder to be used in the
            layer. Default ``None``.

    Examples::

        patch_emb = tensorneko.layer.PatchEmbedding2d(
            image_size=(3, 256, 256),
            patch_size=(16, 16),
            embedding=1024
        )

    References:
        Dosovitskiy, A., Beyer, L., Kolesnikov, A., Weissenborn, D., Zhai, X., Unterthiner, T., ... & Houlsby, N.
        (2020). An image is worth 16x16 words: Transformers for image recognition at scale.
        arXiv preprint arXiv:2010.11929.

    """

    def __init__(self, image_size: Shape, patch_size: Union[int, Shape], embedding: int,
        strides: Optional[Union[int, Shape]] = None,
        build_normalization: Optional[ModuleFactory] = None
    ):
        super().__init__()
        # channel, height, width
        c, h, w = image_size
        # patch_height, patch_width
        ph, pw = (patch_size, patch_size) if type(patch_size) is int else patch_size

        # configure the strides for conv2d
        if strides is None:
            # no specified means no overlap and gap between patches
            strides = (ph, pw)
        elif type(strides) is int:
            # transform the side length of strides to 2D
            strides = (strides, strides)

        self.projection = Conv2d(c, embedding, kernel_size=(ph, pw), stride=strides)
        self.has_norm = build_normalization is not None
        if self.has_norm:
            self.normalization = build_normalization()
        self.rearrange = Rearrange("b d nh nw -> b (nh nw) d")

    def forward(self, x: Tensor) -> Tensor:
        f = F() >> self.projection >> self.rearrange
        if self.has_norm:
            f = f >> self.normalization
        return f(x)



from dataclasses import dataclass

from torch import Tensor


@dataclass
class VideoInfo:
    """A dataclass for info of video"""
    video_fps: float
    audio_fps: int


@dataclass
class VideoData:
    """
    A dataclass for output of :class:`~.video_reader.VideoReader`

    Attributes:
        video (:class:`~torch.Tensor`): video float tensor of (T, C, H, W)
        audio (:class:`~torch.Tensor`): audio float tensor of (C, T)
        info (:class:`VideoInfo`): video and audio fps info
    """
    video: Tensor
    audio: Tensor
    info: VideoInfo

    def __init__(self, video, audio, info):
        self.video = video
        self.audio = audio
        self.info = VideoInfo(**info)



from pytorch_lightning import Callback, Trainer, LightningModule


class DisplayMetricsCallback(Callback):
    """A test callback to display the callback metrics."""

    def on_validation_epoch_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        print(trainer.logger_connector.callback_metrics)


from __future__ import annotations

import os
from datetime import timedelta
from pathlib import Path
from time import time
from typing import Union, Optional, List, Dict

from fn import F
from pytorch_lightning import Callback
from pytorch_lightning import Trainer
from pytorch_lightning.accelerators import Accelerator
from pytorch_lightning.callbacks import ModelCheckpoint
from pytorch_lightning.loggers import TensorBoardLogger
from pytorch_lightning.plugins import Plugin
from pytorch_lightning.plugins.environments import ClusterEnvironment
from pytorch_lightning.profiler import BaseProfiler

from .callback import LrLogger
from .callback.nil_callback import NilCallback


def NekoTrainer(
    logger: Optional[str],
    checkpoint_callback: Union[bool, Callback] = True,
    callbacks: Optional[Union[List[Callback], Callback]] = None,
    default_root_dir: Optional[str] = None,
    gradient_clip_val: float = 0.0,
    gradient_clip_algorithm: str = 'norm',
    process_position: int = 0,
    num_nodes: int = 1,
    num_processes: int = 1,
    gpus: Optional[Union[List[int], str, int]] = None,
    auto_select_gpus: bool = False,
    tpu_cores: Optional[Union[List[int], str, int]] = None,
    log_gpu_memory: Optional[str] = None,
    progress_bar_refresh_rate: Optional[int] = None,
    overfit_batches: Union[int, float] = 0.0,
    track_grad_norm: Union[int, float, str] = -1,
    check_val_every_n_epoch: int = 1,
    fast_dev_run: Union[int, bool] = False,
    accumulate_grad_batches: Union[int, Dict[int, int], List[list]] = 1,
    max_epochs: Optional[int] = None,
    min_epochs: Optional[int] = None,
    max_steps: Optional[int] = None,
    min_steps: Optional[int] = None,
    max_time: Optional[Union[str, timedelta, Dict[str, int]]] = None,
    limit_train_batches: Union[int, float] = 1.0,
    limit_val_batches: Union[int, float] = 1.0,
    limit_test_batches: Union[int, float] = 1.0,
    limit_predict_batches: Union[int, float] = 1.0,
    val_check_interval: Union[int, float] = 1.0,
    flush_logs_every_n_steps: int = 100,
    log_every_n_steps: int = 0,
    accelerator: Optional[Union[str, Accelerator]] = None,
    sync_batchnorm: bool = False,
    precision: int = 32,
    weights_summary: Optional[str] = 'top',
    weights_save_path: Optional[str] = None,
    num_sanity_val_steps: int = 2,
    truncated_bptt_steps: Optional[int] = None,
    resume_from_checkpoint: Optional[Union[Path, str]] = None,
    profiler: Optional[Union[BaseProfiler, str]] = None,
    benchmark: bool = False,
    deterministic: bool = False,
    reload_dataloaders_every_epoch: bool = False,
    auto_lr_find: Union[bool, str] = False,
    replace_sampler_ddp: bool = True,
    terminate_on_nan: bool = False,
    auto_scale_batch_size: Union[str, bool] = False,
    prepare_data_per_node: bool = True,
    plugins: Optional[Union[List[Union[Plugin, ClusterEnvironment, str]], Plugin, ClusterEnvironment, str]] = None,
    amp_backend: str = 'native',
    amp_level: str = 'O2',
    distributed_backend: Optional[str] = None,
    move_metrics_to_cpu: bool = False,
    multiple_trainloader_mode: str = 'max_size_cycle',
    stochastic_weight_avg: bool = False
) -> _NekoTrainer:
    """
    The NekoTrainer wrapper follows the NekoTrainer in PyTorch Lightning, but with some modification in logger, checkpoint,
    and callbacks.

    Two TensorBoard are used for training and validation respectively, for better visual effects in TensorBoard
    visualization. For example, you can see training loss and validation loss in a same line chart. The TensorBoard
    logs are saved in `default_root_dir`/logs.

    Also, the learning rate for each optimizer are logged.

    Args:
        logger (``bool``): Option if you want to use TensorBoard logger in the trainer.

        checkpoint_callback (``bool`` | :class:`~pytorch_lightning.Callback`):
            Set `True` for default checkpoint callback,
            set `False` for no checkpoint callback,
            or a user-defined checkpoint callback object.

        callbacks (``List`` [:class:`~pytorch_lightning.Callback`] | :class:`~pytorch_lightning.Callback`):
            Add a callback or list of callbacks.

        default_root_dir: Default path for logs and weights when no logger/ckpt_callback passed.
            Default: ``os.getcwd()``.
            Can be remote file paths such as `s3://mybucket/path` or 'hdfs://path/'

        gradient_clip_val: 0 means don't clip.

        gradient_clip_algorithm: 'value' means clip_by_value, 'norm' means clip_by_norm. Default: 'norm'

        process_position: orders the progress bar when running multiple models on same machine.

        num_nodes: number of GPU nodes for distributed training.

        num_processes: number of processes for distributed training with distributed_backend="ddp_cpu"

        gpus: number of gpus to train on (int) or which GPUs to train on (list or str) applied per node

        auto_select_gpus: If enabled and `gpus` is an integer, pick available
            gpus automatically. This is especially useful when
            GPUs are configured to be in "exclusive mode", such
            that only one process at a time can access them.

        tpu_cores: How many TPU cores to train on (1 or 8) / Single TPU to train on [1]

        log_gpu_memory: None, 'min_max', 'all'. Might slow performance

        progress_bar_refresh_rate: How often to refresh progress bar (in steps). Value ``0`` disables progress bar.
            Ignored when a custom progress bar is passed to :paramref:`~Trainer.callbacks`. Default: None, means
            a suitable value will be chosen based on the environment (terminal, Google COLAB, etc.).

        overfit_batches: Overfit a fraction of training data (float) or a set number of batches (int).

        track_grad_norm: -1 no tracking. Otherwise tracks that p-norm. May be set to 'inf' infinity-norm.

        check_val_every_n_epoch: Check val every n train epochs.

        fast_dev_run: runs n if set to ``n`` (int) else 1 if set to ``True`` batch(es)
            of train, val and test to find any bugs (ie: a sort of unit test).

        accumulate_grad_batches: Accumulates grads every k batches or as set up in the dict.

        max_epochs: Stop training once this number of epochs is reached. Disabled by default (None).
            If both max_epochs and max_steps are not specified, defaults to ``max_epochs`` = 1000.

        min_epochs: Force training for at least these many epochs. Disabled by default (None).
            If both min_epochs and min_steps are not specified, defaults to ``min_epochs`` = 1.

        max_steps: Stop training after this number of steps. Disabled by default (None).

        min_steps: Force training for at least these number of steps. Disabled by default (None).

        max_time: Stop training after this amount of time has passed. Disabled by default (None).
            The time duration can be specified in the format DD:HH:MM:SS (days, hours, minutes seconds), as a
            :class:`datetime.timedelta`, or a dictionary with keys that will be passed to
            :class:`datetime.timedelta`.

        limit_train_batches: How much of training dataset to check (float = fraction, int = num_batches)

        limit_val_batches: How much of validation dataset to check (float = fraction, int = num_batches)

        limit_test_batches: How much of test dataset to check (float = fraction, int = num_batches)

        limit_predict_batches: How much of prediction dataset to check (float = fraction, int = num_batches)

        val_check_interval: How often to check the validation set. Use float to check within a training epoch,
            use int to check every n steps (batches).

        flush_logs_every_n_steps: How often to flush logs to disk (defaults to every 100 steps).

        log_every_n_steps: Default 0. How often to log within steps (defaults to every 50 steps).
            0 means only log per epoch.

        accelerator: Previously known as distributed_backend (dp, ddp, ddp2, etc...).
            Can also take in an accelerator object for custom hardware.

        sync_batchnorm: Synchronize batch norm layers between process groups/whole world.

        precision: Double precision (64), full precision (32) or half precision (16). Can be used on CPU, GPU or
            TPUs.

        weights_summary: Prints a summary of the weights when training begins.

        weights_save_path: Where to save weights if specified. Will override default_root_dir
            for checkpoints only. Use this if for whatever reason you need the checkpoints
            stored in a different place than the logs written in `default_root_dir`.
            Can be remote file paths such as `s3://mybucket/path` or 'hdfs://path/'
            Defaults to `default_root_dir`.

        num_sanity_val_steps: Sanity check runs n validation batches before starting the training routine.
            Set it to `-1` to run all batches in all validation dataloaders.

        truncated_bptt_steps: Deprecated in v1.3 to be removed in 1.5.
            Please use :paramref:`~pytorch_lightning.core.lightning.LightningModule.truncated_bptt_steps` instead.

        resume_from_checkpoint: Path/URL of the checkpoint from which training is resumed. If there is
            no checkpoint file at the path, start from scratch. If resuming from mid-epoch checkpoint,
            training will start from the beginning of the next epoch.

        profiler: To profile individual steps during training and assist in identifying bottlenecks.

        benchmark: If true enables cudnn.benchmark.

        deterministic: If true enables cudnn.deterministic.

        reload_dataloaders_every_epoch: Set to True to reload dataloaders every epoch.

        auto_lr_find: If set to True, will make trainer.tune() run a learning rate finder,
            trying to optimize initial learning for faster convergence. trainer.tune() method will
            set the suggested learning rate in self.lr or self.learning_rate in the LightningModule.
            To use a different key set a string instead of True with the key name.

        replace_sampler_ddp: Explicitly enables or disables sampler replacement. If not specified this
            will toggled automatically when DDP is used. By default it will add ``shuffle=True`` for
            train sampler and ``shuffle=False`` for val/test sampler. If you want to customize it,
            you can set ``replace_sampler_ddp=False`` and add your own distributed sampler.

        terminate_on_nan: If set to True, will terminate training (by raising a `ValueError`) at the
            end of each training batch, if any of the parameters or the loss are NaN or +/-inf.

        auto_scale_batch_size: If set to True, will `initially` run a batch size
            finder trying to find the largest batch size that fits into memory.
            The result will be stored in self.batch_size in the LightningModule.
            Additionally, can be set to either `power` that estimates the batch size through
            a power search or `binsearch` that estimates the batch size through a binary search.

        prepare_data_per_node: If True, each LOCAL_RANK=0 will call prepare data.
            Otherwise only NODE_RANK=0, LOCAL_RANK=0 will prepare data

        plugins: Plugins allow modification of core behavior like ddp and amp, and enable custom lightning plugins.

        amp_backend: The mixed precision backend to use ("native" or "apex")

        amp_level: The optimization level to use (O1, O2, etc...).

        distributed_backend: deprecated. Please use 'accelerator'

        move_metrics_to_cpu: Whether to force internal logged metrics to be moved to cpu.
            This can save some gpu memory, but can make training slower. Use with attention.

        multiple_trainloader_mode: How to loop over the datasets when there are multiple train loaders.
            In 'max_size_cycle' mode, the trainer ends one epoch when the largest dataset is traversed,
            and smaller datasets reload when running out of their data. In 'min_size' mode, all the datasets
            reload when reaching the minimum length of datasets.

        stochastic_weight_avg: Whether to use `Stochastic Weight Averaging (SWA)
            <https://pytorch.org/blog/pytorch-1.6-now-includes-stochastic-weight-averaging/>_`

    Returns:
        :class:`~._NekoTrainer`: A NekoTrainer object.

    """

    build_trainer = F(_NekoTrainer, logger, checkpoint_callback, callbacks, default_root_dir, gradient_clip_val,
        gradient_clip_algorithm, process_position, num_nodes, num_processes, gpus, auto_select_gpus,
        tpu_cores, log_gpu_memory, progress_bar_refresh_rate, overfit_batches, track_grad_norm,
        check_val_every_n_epoch, fast_dev_run, accumulate_grad_batches, max_epochs, min_epochs,
        max_steps, min_steps, max_time, limit_train_batches, limit_val_batches, limit_test_batches,
        limit_predict_batches, val_check_interval, flush_logs_every_n_steps, log_every_n_steps,
        accelerator, sync_batchnorm, precision, weights_summary, weights_save_path,
        num_sanity_val_steps, truncated_bptt_steps, resume_from_checkpoint, profiler, benchmark,
        deterministic, reload_dataloaders_every_epoch, auto_lr_find, replace_sampler_ddp,
        terminate_on_nan, auto_scale_batch_size, prepare_data_per_node, plugins, amp_backend,
        amp_level, distributed_backend, move_metrics_to_cpu, multiple_trainloader_mode,
        stochastic_weight_avg
    )
    try:
        _NekoTrainer.logger = None
        obj = build_trainer()
        _NekoTrainer.logger = _logger
    except AttributeError:
        obj = build_trainer()
        _NekoTrainer.logger = _logger
    return obj


@property
def _logger(self):
    if self.has_no_logger:
        return None
    if self.training:
        return self.logger_train
    else:
        return self.logger_val


class _NekoTrainer(Trainer):
    """
    The NekoTrainer wrapper follows the NekoTrainer in PyTorch Lightning, but with some modification in logger, checkpoint,
    and callbacks.

    Please use :func:`NekoTrainer` to instantiate a NekoTrainer object.

    Two TensorBoard are used for training and validation respectively, for better visual effects in TensorBoard
    visualization. For example, you can see training loss and validation loss in a same line chart. The TensorBoard
    logs are saved in `default_root_dir`/logs.

    Also, the learning rate for each optimizer are logged.
    """

    def __init__(
        self,
        logger: Optional[str],
        checkpoint_callback: Union[bool, Callback] = True,
        callbacks: Optional[Union[List[Callback], Callback]] = None,
        default_root_dir: Optional[str] = None,
        gradient_clip_val: float = 0.0,
        gradient_clip_algorithm: str = 'norm',
        process_position: int = 0,
        num_nodes: int = 1,
        num_processes: int = 1,
        gpus: Optional[Union[List[int], str, int]] = None,
        auto_select_gpus: bool = False,
        tpu_cores: Optional[Union[List[int], str, int]] = None,
        log_gpu_memory: Optional[str] = None,
        progress_bar_refresh_rate: Optional[int] = None,
        overfit_batches: Union[int, float] = 0.0,
        track_grad_norm: Union[int, float, str] = -1,
        check_val_every_n_epoch: int = 1,
        fast_dev_run: Union[int, bool] = False,
        accumulate_grad_batches: Union[int, Dict[int, int], List[list]] = 1,
        max_epochs: Optional[int] = None,
        min_epochs: Optional[int] = None,
        max_steps: Optional[int] = None,
        min_steps: Optional[int] = None,
        max_time: Optional[Union[str, timedelta, Dict[str, int]]] = None,
        limit_train_batches: Union[int, float] = 1.0,
        limit_val_batches: Union[int, float] = 1.0,
        limit_test_batches: Union[int, float] = 1.0,
        limit_predict_batches: Union[int, float] = 1.0,
        val_check_interval: Union[int, float] = 1.0,
        flush_logs_every_n_steps: int = 100,
        log_every_n_steps: int = 0,  # 0 means for log_every_epoch
        accelerator: Optional[Union[str, Accelerator]] = None,
        sync_batchnorm: bool = False,
        precision: int = 32,
        weights_summary: Optional[str] = 'top',
        weights_save_path: Optional[str] = None,
        num_sanity_val_steps: int = 2,
        truncated_bptt_steps: Optional[int] = None,
        resume_from_checkpoint: Optional[Union[Path, str]] = None,
        profiler: Optional[Union[BaseProfiler, str]] = None,
        benchmark: bool = False,
        deterministic: bool = False,
        reload_dataloaders_every_epoch: bool = False,
        auto_lr_find: Union[bool, str] = False,
        replace_sampler_ddp: bool = True,
        terminate_on_nan: bool = False,
        auto_scale_batch_size: Union[str, bool] = False,
        prepare_data_per_node: bool = True,
        plugins: Optional[
            Union[List[Union[Plugin, ClusterEnvironment, str]], Plugin, ClusterEnvironment, str]] = None,
        amp_backend: str = 'native',
        amp_level: str = 'O2',
        distributed_backend: Optional[str] = None,
        move_metrics_to_cpu: bool = False,
        multiple_trainloader_mode: str = 'max_size_cycle',
        stochastic_weight_avg: bool = False
    ):
        # define a random logger name
        self.log_name = f"{logger}_{int(time())}"
        # build checkpoint callback or from user defined
        if type(checkpoint_callback) is bool and checkpoint_callback is True:
            checkpoint_cb_obj = ModelCheckpoint(dirpath=os.path.join("logs", self.log_name, "checkpoints"),
                save_last=True, filename="{epoch}-{val_loss:.3f}", monitor="val_loss", mode="min"
            )
            checkpoint_cb_bool = True
        elif type(checkpoint_callback) is ModelCheckpoint:
            checkpoint_cb_obj = checkpoint_callback
            checkpoint_cb_bool = True
        elif type(checkpoint_callback) is bool and checkpoint_callback is False:
            checkpoint_cb_obj = NilCallback()
            checkpoint_cb_bool = False
        else:
            raise ValueError("Not available value of checkpoint_callback.")

        # merge the checkpoint callback with other callbacks
        if callbacks is None:
            cbs = [checkpoint_cb_obj]
        elif isinstance(callbacks, Callback):
            cbs = [callbacks, checkpoint_cb_obj]
        else:
            cbs = callbacks + [checkpoint_cb_obj]

        # set learning rate logger
        cbs.append(LrLogger())

        # setup the log mode
        self.log_every_n_steps = log_every_n_steps
        self.log_on_epoch = log_every_n_steps == 0
        self.log_on_step = log_every_n_steps > 0
        if self.log_on_epoch:
            log_every_n_steps = 10000000

        super().__init__(logger is not None, checkpoint_cb_bool, cbs, default_root_dir, gradient_clip_val,
            gradient_clip_algorithm, process_position, num_nodes, num_processes, gpus, auto_select_gpus,
            tpu_cores, log_gpu_memory, progress_bar_refresh_rate, overfit_batches, track_grad_norm,
            check_val_every_n_epoch, fast_dev_run, accumulate_grad_batches, max_epochs, min_epochs,
            max_steps, min_steps, max_time, limit_train_batches, limit_val_batches, limit_test_batches,
            limit_predict_batches, val_check_interval, flush_logs_every_n_steps, log_every_n_steps,
            accelerator, sync_batchnorm, precision, weights_summary, weights_save_path,
            num_sanity_val_steps, truncated_bptt_steps, resume_from_checkpoint, profiler, benchmark,
            deterministic, reload_dataloaders_every_epoch, auto_lr_find, replace_sampler_ddp,
            terminate_on_nan, auto_scale_batch_size, prepare_data_per_node, plugins, amp_backend,
            amp_level, distributed_backend, move_metrics_to_cpu, multiple_trainloader_mode,
            stochastic_weight_avg
        )
        # init loggers for training and validation
        self.has_no_logger = logger is None

        self.logger_train = TensorBoardLogger(save_dir=self.default_root_dir, name="logs",
            version=os.path.join(self.log_name, "train"), log_graph=False  # TODO: Fix log_Graph
        ) if self.has_no_logger is not None else None
        self.logger_val = TensorBoardLogger(save_dir=self.default_root_dir, name="logs",
            version=os.path.join(self.log_name, "val"), log_graph=False
        ) if self.has_no_logger is not None else None



from .audio import AudioReader
from .image import ImageReader
from .text import TextReader
from .video import VideoReader


class read:
    """
    A read util for read different modality of data.

    Examples:

        Read audio:

        .. code-block:: python

            audio: AudioData = read.audio.of(path, return_info=False)

        Read video:

        .. code-block:: python

            video: VideoData = read.video.of(path)

        Read image:

        .. code-block:: python

            image: Tensor = read.image.of(path)

        Read text:

        .. code-block:: python

            text: str = read.text.of(path)

    """

    video = VideoReader
    audio = AudioReader
    text = TextReader
    image = ImageReader



from typing import Union, Sequence, Optional, Callable

from fn import F
from torch import Tensor
from torch.nn import Conv2d as PtConv2d, Module

from ..neko_module import NekoModule


class Conv2d(NekoModule):
    """
    An enhanced Conv2d version of :class:`torch.nn.Conv2d` with combining activation and normalization.

    Args:
        in_channels (``int``): Number of channels in the input image

        out_channels (``int``): Number of channels produced by the convolution

        kernel_size (``int`` | ``tuple``): Size of the convolving kernel

        stride (``int`` | ``tuple``, optional): Stride of the convolution. Default: 1

        padding (``int`` | ``tuple`` | ``str``, optional): Padding added to all four sides of
            the input. Default: 0

        dilation (``int`` | ``tuple``, optional): Spacing between kernel elements. Default: 1

        groups (``int``, optional): Number of blocked connections from input
            channels to output channels. Default: 1

        bias (``bool``, optional): If ``True``, adds a learnable bias to the
            output. Default: ``True``

        padding_mode (``string``, optional): ``'zeros'``, ``'reflect'``,
            ``'replicate'`` or ``'circular'``. Default: ``'zeros'``

        build_activation (``() -> torch.nn.Module``): An activation module builder to be used in the Conv2d layer.

        build_normalization (``() -> torch.nn.Module``): An normalization module builder to be used in the layer.

        normalization_after_activation (``bool``, optional): Set True then applying normalization after activation.
            Default ``False``.

    Attributes:
        conv2d: (:class:`~torch.nn.Conv2d`): The PyTorch Conv2d object.

        activation: (:class:`~torch.nn.Module`): The PyTorch activation module in the layer.

        normalization: (:class:`~torch.nn.Module`): The PyTorch normalization module in the layer.

    Examples::

        conv2d = tensorneko.layer.Conv2d(
            in_channels=256,
            out_channels=1024,
            kernel_size=(3, 3),
            padding=(1, 1),
            build_activation=torch.nn.ReLU,
            build_normalization=lambda: torch.nn.BatchNorm2d(256),
            normalization_after_activation=False
        )

    """

    def __init__(self, in_channels: int, out_channels: int, kernel_size: Union[int, Sequence],
        stride: Union[int, Sequence] = 1, padding: Union[int, Sequence[int]] = 0,
        dilation: Union[int, Sequence] = 1, groups: int = 1, bias: bool = True, padding_mode: str = 'zeros',
        build_activation: Optional[Callable[[], Module]] = None,
        build_normalization: Optional[Callable[[], Module]] = None,
        normalization_after_activation: bool = False
    ):
        super().__init__()

        self.conv2d = PtConv2d(in_channels, out_channels, kernel_size, stride, padding, dilation, groups, bias,
            padding_mode
        )

        self.has_act = build_activation is not None
        if self.has_act:
            self.activation = build_activation()

        self.has_norm = build_normalization is not None
        if self.has_norm:
            self.normalization = build_normalization()
            self.norm_after_act = normalization_after_activation

    def forward(self, x: Tensor) -> Tensor:
        f = F() >> self.conv2d
        if self.has_act and self.has_norm:
            if self.norm_after_act:
                f = f >> self.activation >> self.normalization
            else:
                f = f >> self.normalization >> self.activation
        elif self.has_act and not self.has_norm:
            f = f >> self.activation
        elif not self.has_act and self.has_norm:
            f = f >> self.normalization
        return f(x)



from enum import Enum
from typing import Tuple

import torch
from cleanfid.resize import make_resizer
from einops import rearrange
from fn import F, _
from torch import Tensor, uint8, float32


class ResizeMethod(Enum):
    """Resize methods can be used in :func:`~.resize_image` and :func:`~.resize_video`"""
    BICUBIC = "bicubic"
    BILINEAR = "bilinear"
    NEAREST = "nearest"
    LANCZOS = "lanczos"
    BOX = "box"


def resize_image(tensor: Tensor, size: Tuple[int, int], resize_method: ResizeMethod = ResizeMethod.BICUBIC) -> Tensor:
    """
    Resizing a image to determined size.

    Args:
        tensor (:class:`~torch.Tensor`): Image tensor (C, H, W)
        size ((``int``, ``int``)): Target size (H, W)
        resize_method (:class:`ResizeMethod`, optional): Resize method. Default bicubic.

    Returns:
        :class:`~torch.Tensor`: The resized image.
    """
    h, w = size
    resizer = make_resizer("PIL", True, resize_method.value, (w, h))
    f = F(rearrange, pattern="c h w -> h w c") \
        >> _ * 255 \
        >> (lambda x: x.to(uint8).numpy()) \
        >> resizer \
        >> torch.from_numpy \
        >> F(rearrange, pattern="h w c -> c h w") \
        >> _ / 255 \
        >> (lambda x: x.to(float32))
    return f(tensor)


def resize_video(tensor: Tensor, size: Tuple[int, int], resize_method: ResizeMethod = ResizeMethod.BICUBIC):
    """
    Resizing a video to determined size.

    Args:
        tensor (:class:`~torch.Tensor`): Video tensor (T, C, H, W)
        size ((``int``, ``int``)): Target size (H, W)
        resize_method (:class:`ResizeMethod`, optional): Resize method. Default bicubic.

    Returns:
        :class:`~torch.Tensor`: The resized video.
    """
    image_resizer = F(resize_image, size=size, resize_method=resize_method) \
                    >> F(rearrange, pattern="c h w -> 1 c h w")
    f = F(map, image_resizer) \
        >> list \
        >> torch.vstack
    return f(tensor)



# import from other libraries
from torch.nn import MultiheadAttention as _Attention

Attention = _Attention

# import from modules
from .concatenate import Concatenate
from .conv2d import Conv2d
from .linear import Linear
from .log import Log
from .patching import Patching, PatchEmbedding2d
from .positional_embedding import PositionalEmbedding
from .reshape import Reshape

__all__ = [
    "Attention",
    "Concatenate",
    "Conv2d",
    "Linear",
    "Log",
    "Patching",
    "PatchEmbedding2d",
    "PositionalEmbedding",
    "Reshape",
]



from einops import rearrange
from matplotlib import pyplot as plt
from torch import Tensor


def imshow(image: Tensor, cmap=None, norm=None, aspect=None, interpolation=None,
    alpha=None, vmin=None, vmax=None, origin=None, extent=None, *,
    filternorm=True, filterrad=4.0, resample=None, url=None,
    data=None, **kwargs
):
    """
    Use matplotlib to show image of tensor with shape (C, H, W).

    Except the first parameter, other parameters are from :func:`matplotlib.pyplot.imshow`.

    Args:
        image (:class:`~torch.Tensor`): The image to show.

        cmap: The Colormap instance or registered colormap name used to map scalar data to colors.
            This parameter is ignored for RGB(A) data.

        norm: The Normalize instance used to scale scalar data to the [0, 1] range before mapping to colors using cmap.
            By default, a linear scaling mapping the lowest value to 0 and the highest to 1 is used. This parameter is
            ignored for RGB(A) data.

        aspect: The aspect ratio of the Axes. This parameter is particularly relevant for images since it determines
            whether data pixels are square.

        interpolation: The interpolation method used.

        alpha: The alpha blending value, between 0 (transparent) and 1 (opaque). If alpha is an array,
            the alpha blending values are applied pixel by pixel, and alpha must have the same shape as image.

        vmin, vmax: When using scalar data and no explicit norm, vmin and vmax define the data range that the colormap
            covers. By default, the colormap covers the complete value range of the supplied data. It is deprecated to
            use vmin/vmax when norm is given. When using RGB(A) data, parameters vmin/vmax are ignored.

        origin: Place the [0, 0] index of the array in the upper left or lower left corner of the Axes. The convention
            (the default) 'upper' is typically used for matrices and images.

        extent: The bounding box in data coordinates that the image will fill. The image is stretched individually along
            x and y to fill the box.

        filternorm: A parameter for the antigrain image resize filter (see the antigrain documentation). If filternorm
            is set, the filter normalizes integer values and corrects the rounding errors. It doesn't do anything with
            the source floating point values, it corrects only integers according to the rule of 1.0 which means that
            any sum of pixel weights must be equal to 1.0. So, the filter function must produce a graph of the proper
            shape.

        filterrad: The filter radius for filters that have a radius parameter, i.e. when interpolation is one
            of: 'sinc', 'lanczos' or 'blackman'.
        resample: When True, use a full resampling method. When False, only resample when the output image is larger
            than the input image.

        url: Set the url of the created AxesImage

        **kwargs: Other parameters for :func:`matplotlib.pyplot.imshow`.

    """
    image = rearrange(image, "c h w -> h w c").cpu()
    plt.imshow(image, cmap, norm, aspect, interpolation, alpha, vmin, vmax, origin, extent,
        filternorm=filternorm, filterrad=filterrad, resample=resample, url=url, data=data, **kwargs)
    plt.show()



from . import callback
from . import io
from . import layer
from . import module
from . import optim
from . import preprocess
from . import util
from . import visualization
from .neko_model import NekoModel
from .neko_trainer import NekoTrainer
from .neko_module import NekoModule

__all__ = [
    "callback",
    "io",
    "layer",
    "module",
    "optim",
    "preprocess",
    "util",
    "visualization",
    "NekoModel",
    "NekoTrainer",
    "NekoModule"
]



from typing import Optional

from fn import F
from torch import Tensor
from torch.nn import Linear as PtLinear, Dropout

from ..neko_module import NekoModule
from ..util import ModuleFactory


class Linear(NekoModule):
    """
    An enhanced Linear version of :class:`torch.nn.Linear` with combining activation, normalization and dropout.

    Args:
        in_features (``int``): size of each input sample

        out_features  (``int``): size of each output sample

        bias (``bool``, optional): If set to ``False``, the layer will not learn an additive bias. Default: ``True``

        build_activation (``() -> torch.nn.Module``): An activation module builder to be used in the Conv2d layer.

        build_normalization (``() -> torch.nn.Module``): An normalization module builder to be used in the layer.

        normalization_after_activation (``bool``, optional): Set True then applying normalization after activation.
            Default ``False``.

        dropout_rate (``float``, optional): The dropout rate for this linear layer. 0 means no dropout applied.
            Default ``0``.

    Attributes:
        linear (:class:`~torch.nn.Linear`): The PyTorch Linear object in this layer.

        activation: (:class:`~torch.nn.Module`): The PyTorch activation module in the layer.

        normalization: (:class:`~torch.nn.Module`): The PyTorch normalization module in the layer.

        dropout: (:class:`~torch.nn.Dropout`): The PyTorch Dropout object in this layer.

    Examples::

        linear = tensorneko.layer.Linear(
            in_features=16384,
            out_features=1024,
            build_activation=torch.nn.LeakyReLU,
            build_normalization=lambda: torch.nn.BatchNorm1d(1024),
            dropout_rate=0.5
        )

    """

    def __init__(self, in_features: int, out_features: int, bias: bool = True,
        build_activation: Optional[ModuleFactory] = None,
        build_normalization: Optional[ModuleFactory] = None,
        normalization_after_activation: bool = False,
        dropout_rate: float = 0.
    ):
        super().__init__()
        self.linear = PtLinear(in_features, out_features, bias)

        self.has_act = build_activation is not None
        if self.has_act:
            self.activation = build_activation()
        else:
            self.activation = None

        self.has_norm = build_normalization is not None
        if self.has_norm:
            self.normalization = build_normalization()
            self.norm_after_act = normalization_after_activation
        else:
            self.normalization = None

        self.has_dropout = dropout_rate > 0
        if self.has_dropout:
            self.dropout = Dropout(dropout_rate)

    def forward(self, x: Tensor) -> Tensor:
        f = F() >> self.linear
        if self.has_act and self.has_norm:
            if self.norm_after_act:
                f = f >> self.activation >> self.normalization
            else:
                f = f >> self.normalization >> self.activation
        elif self.has_act and not self.has_norm:
            f = f >> self.activation
        elif not self.has_act and self.has_norm:
            f = f >> self.normalization
        if self.has_dropout:
            f = f >> self.dropout
        return f(x)



from typing import Callable, Union, List, Tuple, TypeVar

import numpy as np
import torch
from torch.nn import Module
from torch import device, Size

ModuleFactory = Union[Callable[[], Module], Callable[[int], Module]]
"""The module builder type of ``() -> torch.nn.Module | (int) -> torch.nn.Module``"""


Device = device
"""Device type of :class:`torch.device`"""


Shape = Union[Size, List[int], Tuple[int, ...]]
"""Shape type. Normally is ``List[int]``, ``Tuple[int, ...]`` or :class:`torch.Size`"""


# Generic types
T = TypeVar('T')  # Any type
A = TypeVar("A", torch.Tensor, np.ndarray)  # Array type


# TODO: Need refactor
from typing import Sequence

from torch import nn
import torch
from ..layer.alae_layers import *


class AlaeStyleEncoderBlock(nn.Module):
    def __init__(self, latent_dim, in_channels, out_channels, downsample=False, is_last_block=False, starting_dim=4,
    ):
        super(AlaeStyleEncoderBlock, self).__init__()
        assert not (is_last_block and downsample), "You should not downscale after last block"
        self.downsample = downsample
        self.is_last_block = is_last_block
        self.conv1 = Lreq_Conv2d(in_channels, in_channels, 3, 1)
        self.lrelu = nn.LeakyReLU(0.2)
        self.instance_norm_1 = StyleInstanceNorm2d(in_channels)
        self.c_1 = LREQ_FC_Layer(2 * in_channels, latent_dim)
        if is_last_block:
            self.conv2 = Lreq_Conv2d(in_channels, out_channels, starting_dim, 0)
            self.c_2 = LREQ_FC_Layer(out_channels, latent_dim)
        else:
            scale = 2 if downsample else 1
            self.conv2 = torch.nn.Sequential(LearnablePreScaleBlur(in_channels),
                                             Lreq_Conv2d(in_channels, out_channels, 3, 1),
                                             torch.nn.AvgPool2d(scale, scale))
            self.instance_norm_2 = StyleInstanceNorm2d(out_channels)
            self.c_2 = LREQ_FC_Layer(2 * out_channels, latent_dim)

        self.name = f"EncodeBlock({latent_dim}, {in_channels}, {out_channels}, is_last_block={is_last_block}, downsample={downsample})"

    def __str__(self):
        return self.name

    def forward(self, x):
        x = self.conv1(x)
        x = self.lrelu(x)

        x, style_1 = self.instance_norm_1(x)
        w1 = self.c_1(style_1.squeeze(3).squeeze(2))

        x = self.conv2(x)
        x = self.lrelu(x)
        if self.is_last_block:
            w2 = self.c_2(x.squeeze(3).squeeze(2))
        else:
            x, style_2 = self.instance_norm_2(x)
            w2 = self.c_2(style_2.squeeze(3).squeeze(2))

        return x, w1, w2


class AlaeEncoder(nn.Module):
    """
    The Style version of ALAE encoder
    """
    def __init__(self, latent_dim, resolutions: Sequence[int], channels: Sequence[int], starting_dim=4):
        """
        progression: A list of tuples (<out_res>, <out_channels>) that describes the Encoding blocks this module should have
        """
        super().__init__()
        self.latent_size = latent_dim
        self.from_rgbs = nn.ModuleList([])
        self.conv_blocks = nn.ModuleList([])

        # Parse the module description given in "progression"
        for i in range(len(resolutions)):
            self.from_rgbs.append(Lreq_Conv2d(3, channels[i], 1, 0))
        for i in range(len(resolutions) - 1, 0, -1):
            if i == 0:
                self.conv_blocks.append(AlaeStyleEncoderBlock(latent_dim, channels[i + 1], channels[0],
                                                              is_last_block=True, starting_dim=starting_dim))
            else:
                downsample = resolutions[i + 1] / 2 == resolutions[i]
                self.conv_blocks.append(AlaeStyleEncoderBlock(latent_dim, channels[i + 1], channels[i],
                                                              downsample=downsample, starting_dim=starting_dim))


        assert(len(self.conv_blocks) == len(self.from_rgbs))
        self.n_layers = len(self.conv_blocks)

    def __str__(self):
        name = "Style-Encoder:\n"
        name += "\tfromRgbs\n"
        for i in range(len(self.from_rgbs)):
            name += f"\t {self.from_rgbs[i]}\n"
        name += "\tStyleEncoderBlocks\n"
        for i in range(len(self.conv_blocks)):
            name += f"\t {self.conv_blocks[i]}\n"
        return name

    def forward(self, image, final_resolution_idx, alpha=1):
        latent_vector = torch.zeros(image.shape[0], self.latent_size).to(image.device)

        feature_maps = self.from_rgbs[final_resolution_idx](image)

        first_layer_idx = self.n_layers - final_resolution_idx - 1
        for i in range(first_layer_idx, self.n_layers):
            feature_maps, w1, w2 = self.conv_blocks[i](feature_maps)
            latent_vector += w1 + w2

            # If this is the first conv block to be run and this is not the last one the there is an already stabilized
            # previous scale layers : Alpha blend the output of the unstable new layer with the downscaled putput
            # of the previous one
            if i == first_layer_idx and i != self.n_layers - 1 and alpha < 1:
                if self.conv_blocks[i].downsample:
                    image = downscale_2d(image)
                skip_first_block_feature_maps = self.from_rgbs[final_resolution_idx - 1](image)
                feature_maps = alpha * feature_maps + (1 - alpha) * skip_first_block_feature_maps

        return latent_vector


# TODO: Need refactor
from typing import Sequence

from torch.nn import Module

from ..layer.alae_layers import *


class StyleGanGenerator(Module):
    """
    The StyleGAN generator.
    """
    def __init__(self, latent_dim, resolutions: Sequence[int], channels: Sequence[int],
        starting_dims: int = 4,
    ):
        """
        style: A list of tuples (<out_res>, <out_channels>) that describes the Generator blocks of this module
        """
        super().__init__()
        self.latent_dim = latent_dim
        self.to_rgb = nn.ModuleList([])
        self.conv_blocks = nn.ModuleList([])

        # Parse the module description given in "style"
        for i in range(len(resolutions)):
            self.to_rgb.append(Lreq_Conv2d(channels[i], 3, 1, 0))
            if i == 0:
                self.conv_blocks.append(StyleGeneratorBlock(latent_dim, channels[0], channels[i + 1],
                                                            is_first_block=True, starting_dims=starting_dims))
            else:
                upscale = (resolutions[i] * 2 == resolutions[i + 1])
                self.conv_blocks.append(StyleGeneratorBlock(latent_dim, channels[i], channels[i + 1],
                                                            upscale=upscale, starting_dims=starting_dims))

    def __str__(self):
        name = "Style-Generator:\n"
        name += "\ttoRgbs\n"
        for i in range(len(self.conv_blocks)):
            name += f"\t {self.to_rgb[i]}\n"
        name += "\tStyleGeneratorBlocks\n"
        for i in range(len(self.conv_blocks)):
            name += f"\t {self.conv_blocks[i]}\n"
        return name

    def forward(self, w, final_resolution_idx, alpha):
        generated_img = None
        feature_maps = None
        for i, block in enumerate(self.conv_blocks):
            # Separate noise for each block
            noise = torch.randn((w.shape[0], 1, 1, 1), dtype=torch.float32).to(w.device)

            prev_feature_maps = feature_maps
            feature_maps = block(feature_maps, w, noise)

            if i == final_resolution_idx:
                generated_img = self.to_rgb[i](feature_maps)

                # If there is an already stabilized last previous resolution layer. alpha blend with it
                if i > 0 and alpha < 1:
                    generated_img_without_last_block = self.to_rgb[i - 1](prev_feature_maps)
                    if block.upscale:
                        generated_img_without_last_block = upscale_2d(generated_img_without_last_block)
                    generated_img = alpha * generated_img + (1 - alpha) * generated_img_without_last_block
                break

        return generated_img


class StyleGeneratorBlock(Module):
    def __init__(self, latent_dim, in_channels, out_channels, is_first_block=False, upscale=False, starting_dims=4):
        super().__init__()
        assert not (is_first_block and upscale), "You should not upscale if this is the first block in the generator"
        self.is_first_block = is_first_block
        self.upscale = upscale
        if is_first_block:
            self.const_input = nn.Parameter(torch.randn(1, out_channels, starting_dims, starting_dims))
        else:
            self.blur = LearnablePreScaleBlur(out_channels)
            self.conv1 = Lreq_Conv2d(in_channels, out_channels, 3, padding=1)

        self.style_affine_transform_1 = StyleAffineTransform(latent_dim, out_channels)
        self.style_affine_transform_2 = StyleAffineTransform(latent_dim, out_channels)
        self.noise_scaler_1 = NoiseScaler(out_channels)
        self.noise_scaler_2 = NoiseScaler(out_channels)
        self.adain = AdaIn(in_channels)
        self.lrelu = nn.LeakyReLU(0.2)
        self.conv2 = Lreq_Conv2d(out_channels, out_channels, 3, padding=1)

        self.name = f"StyleBlock({latent_dim}, {in_channels}, {out_channels}, is_first_block={is_first_block}, upscale={upscale})"

    def __str__(self):
        return self. name

    def forward(self, input, latent_w, noise):
        if self.is_first_block:
            assert(input is None)
            result = self.const_input.repeat(latent_w.shape[0], 1, 1, 1)
        else:
            if self.upscale:
                input = upscale_2d(input)
            result = self.conv1(input)
            result = self.blur(result)

        result += self.noise_scaler_1(noise)
        result = self.adain(result, self.style_affine_transform_1(latent_w))
        result = self.lrelu(result)

        result = self.conv2(result)
        result += self.noise_scaler_2(noise)
        result = self.adain(result, self.style_affine_transform_2(latent_w))
        result = self.lrelu(result)

        return result


from typing import Iterable

from torch import Tensor
from torch.nn import Sequential, ModuleList

from ..neko_module import NekoModule
from ..layer import Concatenate
from ..util import ModuleFactory


class DenseBlock(NekoModule):
    """
    The DenseBlock can be used to build a block with repeatable submodules with dense connections. This structure is
    proposed by Huang, Liu, Van Der Maaten, & Weinberger (2017).

    Args:
        sub_module_layers (``List`` [``(int) -> torch.nn.Module``]):
            A collection of module factory builder to build a "layer" in DenseBlock. In the DenseBlock, there will be a
            submodule generated repeatedly for several times. The factory function takes an repeat_index as input, and
            build a :class:`~torch.nn.Module`.

        repeat (``int``): Number of repeats for each layer in DenseBlock.

    Attributes:
        build_sub_module (``(int) -> torch.nn.Module``): The module factory function to build a submodule in DenseBlock.

        sub_modules (:class:`~torch.nn.ModuleList`): The ModuleList of all submodules.

        concatenates (:class:`~torch.nn.ModuleList`): The ModuleList of all Concatenate layers in DenseBlock.

    Examples::

        # batch norm builder
        def build_bn(i=0):
            return BatchNorm2d(2 ** i * self.c)

        # conv2d builder
        def build_conv2d_1x1(i=0):
            return Conv2d(2 ** i * self.c, 2 ** i * self.c * 4, (1, 1), build_activation=ReLU,
                build_normalization=lambda: BatchNorm2d(2 ** i * self.c * 4))

        # conv
        def build_conv2d_3x3(i=0):
            return Conv2d(2 ** i * self.c * 4, 2 ** i * self.c, (3, 3), padding=(1, 1))

        dense_block = tensorneko.module.DenseBlock((
            build_bn,
            lambda i: ReLU(),
            build_conv2d_1x1,
            build_conv2d_3x3
        ), repeat=4)

    References:
        Huang, G., Liu, Z., Van Der Maaten, L., & Weinberger, K. Q. (2017). Densely connected convolutional networks.
        In Proceedings of the IEEE conference on computer vision and pattern recognition (pp. 4700-4708).

    """

    def __init__(self, sub_module_layers: Iterable[ModuleFactory], repeat: int = 2):
        super().__init__()

        def build_sub_module(i):
            return Sequential(*[
                build_layer(i) for build_layer in sub_module_layers
            ])

        self.build_sub_module = build_sub_module
        self.sub_modules = ModuleList([
            self.build_sub_module(i) for i in range(repeat)
        ])
        self.concatenates = ModuleList([
            Concatenate(dim=1) for i in range(repeat)
        ])

    def forward(self, x: Tensor) -> Tensor:
        xs = []
        for i in range(len(self.sub_modules)):
            # concat with previous output
            xs.append(x)
            if i != 0:
                x = self.concatenates[i - 1](xs)
            # forward with submodule
            x = self.sub_modules[i](x)

        xs.append(x)
        x = self.concatenates[-1](xs)
        return x



from .display_metrics_callback import DisplayMetricsCallback
from .lr_logger import LrLogger
from .nil_callback import NilCallback

__all__ = [
    "LrLogger",
    "NilCallback",
    "DisplayMetricsCallback"
]



from typing import Optional, Callable, Sequence

from torch import Tensor
from torch.nn import Module, ModuleList, Sequential

from ..neko_module import NekoModule
from ..util import compose, ModuleFactory


class ResidualBlock(NekoModule):
    """
    The ResidualBlock module is a single block of a residual connected block.
    x -> sub_module -> add_to(x) -> tail_module

    The residual connection is introduced firstly by He, Zhang, Ren and Sun (2016).

    Args:
        sub_module_layers (`Sequence[() -> Module]`): A collection of module builders for submodule route.

        tail_module_layers (`Sequence[() -> Module]`, optional): A collection of module builders for tail module which
            is placed after residual addition.

    Attributes:
        sub_module (:class:`~torch.nn.Sequential`): The PyTorch Sequential object as submodule route.

        tail_module (:class:`~torch.nn.Sequential` | ``None``): The PyTorch Sequential object as tail module after
            residual addition.

    Examples::

        block = ResidualBlock([
            lambda: torch.nn.Linear(24, 48),
            torch.nn.ReLU,
            lambda: torch.nn.Linear(48, 24)
        ])

    References:
        He, K., Zhang, X., Ren, S., & Sun, J. (2016). Deep residual learning for image recognition. In Proceedings of
        the IEEE conference on computer vision and pattern recognition (pp. 770-778).

    """

    def __init__(self, sub_module_layers: Sequence[ModuleFactory],
        tail_module_layers: Optional[Sequence[ModuleFactory]] = None
    ):

        super().__init__()
        self.sub_module = Sequential(*[build_layer() for build_layer in sub_module_layers])
        self.tail_module = None if tail_module_layers is None else \
            Sequential(*[build_layer() for build_layer in tail_module_layers])

    def forward(self, x: Tensor) -> Tensor:
        x = self.sub_module(x) + x
        return self.tail_module(x) if self.tail_module else x


class ResidualModule(NekoModule):
    """
    The Residual Module repeatedly generate Residual Blocks.

    Args:
        build_block (``() -> ResidualBlock``): The ResidualBlock builder function for repeatedly building.

        repeat (``int``): The repeat number of residual block.

    Attributes:
        blocks (:class:`~torch.nn.ModuleList`): The blocks list in the module.

    Examples::

        def build_block():
            return ResidualBlock([
                lambda: torch.nn.Linear(24, 48),
                torch.nn.ReLU,
                lambda: torch.nn.Linear(48, 24)
            ])

        residual_module = ResidualModule(build_block, repeat=10)

    """

    def __init__(self, build_block: Callable[[], ResidualBlock], repeat: int):
        super().__init__()
        self.blocks = ModuleList([build_block() for _ in range(repeat)])
        self.repeat = repeat

    def forward(self, x: Tensor) -> Tensor:
        f = compose(self.blocks)
        return f(x)



from torch import Tensor
from torchvision.io import read_image
from torchvision.io.image import ImageReadMode


class ImageReader:
    """ImageReader for reading images as :class:`~torch.Tensor`"""

    @staticmethod
    def of(path: str, mode: ImageReadMode = ImageReadMode.UNCHANGED) -> Tensor:
        """
        Read image tensor of given file.

        Args:
            path (``str``): Path of the JPEG or PNG image.
            mode (:class:`~torchvision.io.image.ImageReadMode`, optional):
                the read mode used for optionally converting the image.
                Default: :class:`~torchvision.io.image.ImageReadMode.UNCHANGED`.
                See :class:`~torchvision.io.image.ImageReadMode` class for more information on various
                available modes.

        Returns:
            :class:`~torch.Tensor`: A float tensor of image (C, H, W), with value range of 0. to 1.
        """
        return read_image(path, mode) / 255



from typing import Dict

from torch.nn import LeakyReLU, GELU, ELU, ReLU, CrossEntropyLoss, L1Loss, MSELoss, BCELoss, BCEWithLogitsLoss, Mish, \
    Module


class StringGetter:
    """Util class for retrieve module class from string"""

    _activation_mapping: Dict[str, Module] = {
        "LEAKYRELU": LeakyReLU,
        "GELU": GELU,
        "ELU": ELU,
        "RELU": ReLU,
        "MISH": Mish,
    }

    _loss_mapping: Dict[str, Module] = {
        "CROSSENTROPYLOSS": CrossEntropyLoss,
        "L1LOSS": L1Loss,
        "MSELOSS": MSELoss,
        "BCELOSS": BCELoss,
        "BCEWITHLOGITSLOSS": BCEWithLogitsLoss
    }

    _mapping: Dict[str, Dict[str, Module]] = {
        "activation": _activation_mapping,
        "loss": _loss_mapping
    }

    def __init__(self, type: str):
        self.type = type
        self.dict = StringGetter._mapping[self.type]

    def __call__(self, name: str):
        return self._get(name)

    def _get(self, name: str):
        return self.dict[name.upper()]


get_activation = StringGetter("activation")
"""The function for getting activation function from its name"""


get_loss = StringGetter("loss")
"""The function for getting loss function from its name"""



from .dense import DenseBlock
from .inception import InceptionModule
from .mlp import MLP
from .residual import ResidualModule, ResidualBlock
from .transformer import TransformerEncoderBlock, AttentionModule, TransformerEncoder

__all__ = [
    "DenseBlock",
    "InceptionModule",
    "MLP",
    "ResidualBlock",
    "ResidualModule",
    "TransformerEncoderBlock",
    "AttentionModule"
]



from fn import F, _
from torch import Tensor, zeros
from torch.nn import Parameter, Dropout

from ..neko_module import NekoModule
from ..util import Shape


class PositionalEmbedding(NekoModule):
    """
    Trainable positional embedding for transformer/self-attention layer.

    The trainable positional embedding is used in Transformer (Vaswani, et al., 2017).

    Args:
        input_shape (:class:`~tensorneko.util.type.Shape`): The input shape of the sequence.

        dropout_rate (``int``, optional): The dropout rate after the positional embedding. Default ``0``.

    Attributes:
        emb (:class`~torch.nn.Parameter`): The trainable positional embedding parameters.

        dropout (:class`~torch.nn.Dropout`): The PyTorch dropout module in the layer.

    References:
        Vaswani, A., Shazeer, N., Parmar, N., Uszkoreit, J., Jones, L., Gomez, A. N., ... & Polosukhin, I. (2017).
        Attention is all you need. In Advances in neural information processing systems (pp. 5998-6008).

    """

    def __init__(self, input_shape: Shape, dropout_rate: float = 0.5):
        super().__init__()
        self.emb = Parameter(zeros(1, *input_shape))
        self.use_dropout = dropout_rate is not None and dropout_rate != 0.
        if self.use_dropout:
            self.dropout = Dropout(dropout_rate)


    def forward(self, x: Tensor) -> Tensor:
        f = F() >> _ + self.emb
        if self.use_dropout:
            f = f >> self.dropout
        return f(x)
